@@ -1,7 +1,9 @@
 # -*- coding: utf8 -*-
 from __future__ import unicode_literals
 from abc import abstractmethod, ABCMeta
+from build.lib import subfinder
 import os
+from os.path import join
 import re
 import cgi
 try:
@@ -55,18 +57,27 @@ class BaseSubSearcher(object):
 
     SUPPORT_LANGUAGES = []
     SUPPORT_EXTS = []
+    shortname = 'base_subsearcher'
+    API_URL = ''
 
-    def __init__(self, subfinder,  **kwargs):
+    def __init__(self, subfinder,  api_urls=None, **kwargs):
         """
         subfinder: SubFinder
-        debug: 是否输出调试信息
+        api_urls: api_urls
         """
         self.session = requests.session()
         self.session.headers['User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36'
         self.subfinder = subfinder
-        self.api_urls = kwargs.get('api_urls', {})
-        self.API_URL = self.api_urls.get(
-            self.shortname, self.__class__.API_URL)
+        self.api_urls = api_urls if api_urls else {}
+        self.API_URL = self.api_urls.get(self.shortname, self.__class__.API_URL)
+
+        self.languages = ''
+        self.exts = ''
+        self.videofile = ''
+        self.videoname = ''
+        self.videoinfo = {}
+        self.keywords = []
+        self.referer = ''
 
     def _debug(self, msg):
         self.subfinder.logger.debug(msg)
@@ -91,6 +102,72 @@ class BaseSubSearcher(object):
         """
         return urlparse.urljoin(url, path)
 
+    @abstractmethod
+    def search_subs(self, videofile, languages=None, exts=None, keyword=None):
+        """ search subtitles of videofile.
+
+        `videofile` is the absolute(or relative) path of the video file.
+
+        `languages` is the language of subtitle, e.g chn, eng, the support for language is difference, depende on
+        implemention of subclass. `languages` accepts one language or a list of language
+
+        `exts` is the format of subtitle, e.g ass, srt, sub, idx, the support for ext is difference,
+        depende on implemention of subclass. `ext` accepts one ext or a list of ext
+
+        `keyword` is used to searching on the subtitle website.
+
+        return a list of subtitle info
+        [
+            {
+                'link': '',     # download link
+                'language': '', # language
+                'ext': '',      # ext
+                'subname': '',  # the filename of subtitles
+                'downloaded': False, # it's tell `SubFinder` whether need to download.
+            },
+            {
+                'link': '',
+                'language': '',
+                'ext': '',
+                'subname': '',
+            },
+            ...
+        ]
+        - `link`, it's optional, but if `downloaded` is False, then `link` is required.
+        - `language`, it's optional
+        - `subname`, it's optional, but if `downloaded` is False, then `subname` is required.
+        - `downloaded`, `downloaded` is required.
+            if `downloaded` is True, then `SubFinder` will not download again,
+            otherwise `SubFinder` will download link.
+        """
+        pass
+
+    def __str__(self):
+        if hasattr(self.__class__, 'shortname'):
+            name = self.__class__.shortname
+        else:
+            name = self.__class__.__name__
+        return '<{}>'.format(name)
+
+    def __unicode__(self):
+        return self.__str__()
+
+
+class HTMLSubSearcher(BaseSubSearcher):
+    __metaclass__ = ABCMeta
+
+    LANGUAGES_MAP = {
+        '简体': 'zh_chs',
+        '繁體': 'zh_cht',
+        'English': 'en',
+        'english': 'en',
+        '英文': 'en',
+        '双语': 'zh_en',
+        '中英': 'zh_en',
+    }
+    COMMON_LANGUAGES = ['简体', '繁体', '英文']
+    shortname = 'html_subsearcher'
+
     @classmethod
     def _get_videoname(cls, videofile):
         """parse the `videofile` and return it's basename
@@ -118,19 +195,10 @@ class BaseSubSearcher(object):
         - sub_title, the sub_title of video
         - resolution,
         - source,
-        -
         - season, defaults to 0
         - episode, defaults to 0
         """
-        info = {
-            'title': '',
-            'season': 0,
-            'episode': 0,
-            'resolution': '',
-            'source': '',
-            'audio_encoding': '',
-            'video_encoding': '',
-        }
+        info = VideoInfo()
         mapping = {
             'resolution': cls.RE_RESOLUTION,
             'source': cls.RE_SOURCE,
@@ -150,7 +218,7 @@ class BaseSubSearcher(object):
                 info['season'] = int(m.group('season'))
                 index, _ = m.span()
                 info['title'] = videoname[0:index].strip('.')
-        
+
         for k, r in mapping.items():
             m = r.search(videoname)
             if m:
@@ -166,22 +234,30 @@ class BaseSubSearcher(object):
 
         return info
 
-    @classmethod
-    def _gen_subname(cls, videofile, language, ext, **kwargs):
-        """ 生成字幕文件名
-        """
-        root, basename = os.path.split(videofile)
-        name, _ = os.path.splitext(basename)
-        subname = '{basename}.{language}.{ext}'.format(
-            basename=name,
-            language=language,
-            ext=ext)
-        return subname
+    def _gen_subname(self, origin_file, language=None, ext=None):
+        if not language:
+            language_ = []
+            try:
+                for l in self.COMMON_LANGUAGES:
+                    if origin_file.find(l) >= 0:
+                        language_.append(l)
+            except Exception:
+                pass
+            language = '&'.join(language_)
+        if language and not language.startswith('.'):
+            language = '.' + language
 
-    @classmethod
-    def _extract(cls, compressed_file, videofile, exts):
+        basename = os.path.basename(self.videofile)
+        basename, _ = os.path.splitext(basename)
+        if not ext:
+            _, ext = os.path.splitext(origin_file)
+        if not ext.startswith('.'):
+            ext = '.' + ext
+
+        return '{basename}{language}{ext}'.format(basename=basename, language=language, ext=ext)
+
+    def _extract(self, compressed_file):
         """ 解压字幕文件，如果无法解压，则直接返回 compressed_file。
-        exts 参数用于过滤掉非字幕文件，只有文件的扩展名在 exts 中，才解压该文件。
         """
         if not CompressedFile.is_compressed_file(compressed_file):
             return [compressed_file]
@@ -193,20 +269,76 @@ class BaseSubSearcher(object):
             if cf.isdir(name):
                 continue
             # make `name` to unicode string
-            orig_name = CompressedFile.decode_file_name(name)
-            _, ext = os.path.splitext(orig_name)
+            origin_file = CompressedFile.decode_file_name(name)
+            _, ext = os.path.splitext(origin_file)
             ext = ext[1:]
-            if ext not in exts:
+            if self.exts and ext not in self.exts:
                 continue
-            subname = cls._gen_subname(videofile, '', '', orig_name=orig_name)
+            subname = self._gen_subname(origin_file)
             subpath = os.path.join(root, subname)
             cf.extract(name, subpath)
             subs.append(subpath)
         cf.close()
         return subs
 
+    def _download_subtitle(self, download_link, subtitle):
+        """ 下载字幕
+        videofile: 视频文件路径
+        sub_title: 字幕标题（文件名）
+        download_link: 下载链接
+        """
+        root = os.path.dirname(self.videofile)
+        name = self.videoname
+        ext = ''
+        res = self.session.get(download_link, headers={'Referer': self.referer}, stream=True)
+        self.referer = res.url
+        # 尝试从 Content-Disposition 中获取文件后缀名
+        content_disposition = res.headers.get('Content-Disposition', '')
+        if content_disposition:
+            _, params = cgi.parse_header(content_disposition)
+            filename = params.get('filename')
+            if filename:
+                _, ext = os.path.splitext(filename)
+                ext = ext[1:]
+        if ext == '':
+            # 尝试从url 中获取文件后缀名
+            p = urlparse.urlparse(res.url)
+            path = p.path
+            if path:
+                _, ext = os.path.splitext(path)
+                ext = ext[1:]
+        if ext == '':
+            # 尝试从字幕标题中获取文件后缀名
+            _, ext = os.path.splitext(subtitle)
+            ext = ext[1:]
+
+        filename = '{}.{}'.format(name, ext)
+        filepath = os.path.join(root, filename)
+        with open(filepath, 'wb') as fp:
+            for chunk in res.iter_content(8192):
+                fp.write(chunk)
+        res.close()
+        return filepath
+
     @classmethod
-    def _filter_subinfo_list(cls, subinfo_list, videoinfo, languages, exts):
+    def _gen_keyword(cls, videoinfo):
+        """ 获取关键词
+        """
+        separators = ['.', ' ']
+        keywords = []
+        for sep in separators:
+            keyword = [videoinfo.get('title')]
+            if videoinfo['season'] != 0:
+                keyword.append('S{:02d}'.format(videoinfo['season']))
+            if videoinfo['episode'] != 0:
+                keyword.append('E{:02d}'.format(videoinfo['episode']))
+            # replace space with ''
+            keyword_str = sep.join(keyword)
+            keyword_str = re.sub(r'\s+', ' ', keyword_str)
+            keywords.append(keyword_str)
+        return keywords
+
+    def _filter_subinfo_list(self, subinfo_list):
         """ filter subinfo list base on:
         - season
         - episode
@@ -223,11 +355,19 @@ class BaseSubSearcher(object):
             'video_encoding',
             'audio_encoding'
         ]
+        videoinfo = self.videoinfo
         filtered_subinfo_list = dict((f, []) for f in filter_field_list)
-
+        languages = set(self.languages)
+        exts = set(self.exts)
         for subinfo in subinfo_list:
+            languages_ = subinfo.get('languages')
+            exts_ = subinfo.get('exts')
+            if languages_ and not (languages & set(languages_)):
+                continue
+            if exts_ and not (exts & set(exts_)):
+                continue
             title = subinfo.get('title')
-            videoinfo_ = cls._parse_videoname(title)
+            videoinfo_ = self._parse_videoname(title)
             last_field = None
             for field in filter_field_list:
                 i = videoinfo.get(field)
@@ -240,105 +380,33 @@ class BaseSubSearcher(object):
                     last_field = field
                 else:
                     break
-            if last_field is not None:
+            if last_field is not None and last_field != filter_field_list[0]:
                 filtered_subinfo_list[last_field].append(subinfo)
-        for field in filter_field_list[::-1]:
+        for field in reversed(filter_field_list):
             if len(filtered_subinfo_list[field]) > 0:
                 # sort by download_count and rate
                 sorted_subinfo_list = sorted(filtered_subinfo_list[field],
-                                             key=lambda item: (
-                    item['rate'], item['download_count']),
-                    reverse=True)
+                                             key=lambda item: (item['rate'], item['download_count']), reverse=True)
                 return sorted_subinfo_list[0]
-        return None
 
-    def _download_subs(self, download_link, videofile, referer='', sub_title=''):
-        """ 下载字幕
-        videofile: 视频文件路径
-        sub_title: 字幕标题（文件名）
-        download_link: 下载链接
-        referer: referer
+    @abstractmethod
+    def _get_subinfo_list(self, keyword):
+        """ return subinfo_list of videoname
         """
-        root = os.path.dirname(videofile)
-        name, _ = os.path.splitext(os.path.basename(videofile))
-        ext = ''
+        # searching subtitles
+        pass
 
-        headers = {
-            'Referer': referer
-        }
-        res = self.session.get(download_link, headers=headers, stream=True)
-        referer = res.url
+    @abstractmethod
+    def _visit_detailpage(self, detailpage_link):
+        pass
 
-        # 尝试从 Content-Disposition 中获取文件后缀名
-        content_disposition = res.headers.get('Content-Disposition', '')
-        if content_disposition:
-            _, params = cgi.parse_header(content_disposition)
-            filename = params.get('filename')
-            if filename:
-                _, ext = os.path.splitext(filename)
-                ext = ext[1:]
-
-        if ext == '':
-            # 尝试从url 中获取文件后缀名
-            p = urlparse.urlparse(res.url)
-            path = p.path
-            if path:
-                _, ext = os.path.splitext(path)
-                ext = ext[1:]
-
-        if ext == '':
-            # 尝试从字幕标题中获取文件后缀名
-            _, ext = os.path.splitext(sub_title)
-            ext = ext[1:]
-
-        filename = '{}.{}'.format(name, ext)
-        filepath = os.path.join(root, filename)
-        with open(filepath, 'wb') as fp:
-            for chunk in res.iter_content(8192):
-                fp.write(chunk)
-
-        return filepath, referer
-
-    def _search_subs(self, videofile, languages, exts):
-        """ search subtitles of videofile.
-
-        `videofile` is the absolute(or relative) path of the video file.
-
-        `languages` is the language of subtitle, e.g chn, eng, the support for language is difference, depende on
-        implemention of subclass. `languages` accepts one language or a list of language
-
-        `exts` is the format of subtitle, e.g ass, srt, sub, idx, the support for ext is difference,
-        depende on implemention of subclass. `ext` accepts one ext or a list of ext
-
-        return a list of subtitle info
-        [
-            {
-                'link': '',     # download link
-                'language': '', # language
-                'ext': '',      # ext
-                'subname': '',  # the filename of subtitles
-                'downloaded': False, # it's tell `SubFinder` whether need to download.
-            },
-            {
-                'link': '',
-                'language': '',
-                'ext': '',
-                'subname': '',
-            },
-            ...
-        ]
-        - `link`, it's optional, but if `downloaded` is False, then `link` is required.
-        - `language`, it's optional
-        - `subname`, it's optional, but if `downloaded` is False, then `subname` is required.
-        - `downloaded`, `downloaded` is required.
-            if `downloaded` is True, then `SubFinder` will not download again,
-            otherwise `SubFinder` will download link.
-
-        sub-class should implement this private method.
+    @abstractmethod
+    def _visit_downloadpage(self, downloadpage_link):
+        """ get the real download link of subtitles.
         """
-        return []
+        pass
 
-    def search_subs(self, videofile, languages=None, exts=None):
+    def _prepare_search_subs(self, videofile, languages=None, exts=None, keyword=None):
         if languages is None:
             languages = self.SUPPORT_LANGUAGES
         elif isinstance(languages, str):
@@ -351,14 +419,102 @@ class BaseSubSearcher(object):
             exts = [exts]
         self._check_exts(exts)
 
-        return self._search_subs(videofile, languages, exts)
-
-    def __str__(self):
-        if hasattr(self.__class__, 'shortname'):
-            name = self.__class__.shortname
+        self.languages = languages
+        self.exts = exts
+        self.videofile = videofile
+        self.videoname = self._get_videoname(videofile)  # basename, not include ext
+        self.videoinfo = self._parse_videoname(self.videoname)
+        if keyword is None:
+            keywords = self._gen_keyword(self.videoinfo)
         else:
-            name = self.__class__.__name__
-        return '<{}>'.format(name)
+            keywords = [keyword]
+        self.keywords = keywords
+    
+    def search_subs(self, videofile, languages=None, exts=None, keyword=None):
+        self._prepare_search_subs(videofile, languages, exts, keyword)
+        self._debug('keywords: {}'.format(self.keywords))
+        self._debug('videoinfo: {}'.format(self.videoinfo))
+        subinfo = self._get_subinfo()
+        if not subinfo:
+            return []
+        subs = self._download_subs(subinfo)
+        return [{
+            'link': self.referer,
+            'language': subinfo['languages'],
+            'ext': subinfo['exts'],
+            'subname': subs,
+            'downloaded': True
+        }] if subs else []
 
-    def __unicode__(self):
-        return self.__str__()
+    def _get_subinfo(self):
+        subinfo = None
+        for keyword in self.keywords:
+            subinfo_list = self._get_subinfo_list(keyword)
+            self._debug('subinfo_list: {}'.format(subinfo_list))
+            subinfo = self._filter_subinfo_list(subinfo_list)
+            self._debug('subinfo: {}'.format(subinfo))
+            if subinfo:
+                break
+        return subinfo
+
+    def _download_subs(self, subinfo):
+        downloadpage_link = self._visit_detailpage(subinfo['link'])
+        self._debug('downloadpage_link: {}'.format(downloadpage_link))
+        subtitle_download_link = self._visit_downloadpage(downloadpage_link)
+        self._debug('subtitle_download_link: {}'.format(subtitle_download_link))
+        filepath = self._download_subtitle(subtitle_download_link, subinfo['title'])
+        self._debug('filepath: {}'.format(filepath))
+        subs = self._extract(filepath)
+        self._debug('subs: {}'.format(subs))
+        return subs
+
+
+class VideoInfo(dict):
+    FIELDS = [
+        'title',
+        'season',
+        'episode',
+        'resolution',
+        'source',
+        'audio_encoding',
+        'video_encoding',
+    ]
+
+    def __init__(self, *args, **kwargs) -> None:
+        super(VideoInfo, self).__init__(*args, **kwargs)
+        self['title'] = ''
+        self['season'] = 0
+        self['episode'] = 0
+        self['resolution'] = ''
+        self['source'] = ''
+        self['audio_encoding'] = ''
+        self['video_encoding'] = ''
+        for k, v in kwargs:
+            if k in self.FIELDS:
+                self[k] = v
+
+
+class SubInfo(dict):
+    FIELDS = [
+        'title',
+        'link',
+        'author',
+        'exts',
+        'languages',
+        'rate',
+        'download_count',
+    ]
+
+    def __init__(self, *args, **kwargs) -> None:
+        super(SubInfo, self).__init__(*args, **kwargs)
+        self['title'] = ''
+        self['link'] = ''
+        self['author'] = ''
+        self['exts'] = []
+        self['languages'] = []
+        self['rate'] = 0
+        self['download_count'] = 0
+
+        for k, v in kwargs:
+            if k in self.FIELDS:
+                self[k] = v
